@@ -1,5 +1,6 @@
 // Copyright 2016 The TensorFlow Authors. All Rights Reserved.
 // Modifications copyright (C) 2019 Uber Technologies, Inc.
+// Modifications copyright (C) 2020, NVIDIA CORPORATION. All rights reserved.
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -41,8 +42,9 @@ ncclDataType_t GetNCCLDataType(const std::shared_ptr<Tensor> tensor) {
   }
 }
 
-void NCCLContext::ErrorCheck(std::string op_name, ncclResult_t nccl_result) {
+void NCCLContext::ErrorCheck(std::string op_name, ncclResult_t nccl_result, ncclComm_t& nccl_comm) {
   if (nccl_result != ncclSuccess) {
+    ncclCommAbort(nccl_comm);
     throw std::logic_error(std::string(op_name) + " failed: " + ncclGetErrorString(nccl_result));
   }
 }
@@ -70,7 +72,7 @@ void NCCLOpContext::InitNCCLComm(const std::vector<TensorTableEntry>& entries,
 
     ncclUniqueId nccl_id;
     if (nccl_rank == 0) {
-      nccl_context_->ErrorCheck("ncclGetUniqueId", ncclGetUniqueId(&nccl_id));
+      nccl_context_->ErrorCheck("ncclGetUniqueId", ncclGetUniqueId(&nccl_id), nccl_comm);
     }
 
     global_state_->controller->Bcast((void*)&nccl_id, sizeof(nccl_id), 0,
@@ -78,7 +80,7 @@ void NCCLOpContext::InitNCCLComm(const std::vector<TensorTableEntry>& entries,
 
     ncclComm_t new_nccl_comm;
     auto nccl_result = ncclCommInitRank(&new_nccl_comm, nccl_size, nccl_id, nccl_rank);
-    nccl_context_->ErrorCheck("ncclCommInitRank", nccl_result);
+    nccl_context_->ErrorCheck("ncclCommInitRank", nccl_result, nccl_comm);
     nccl_comm = new_nccl_comm;
 
     // Barrier helps NCCL to synchronize after initialization and avoid
@@ -91,6 +93,21 @@ void NCCLOpContext::InitNCCLComm(const std::vector<TensorTableEntry>& entries,
   nccl_comm_ = &nccl_comm;
 }
 
+void NCCLOpContext::AsyncErrorCheck() {
+  ncclResult_t nccl_async_err;
+  auto nccl_err = ncclCommGetAsyncError(*nccl_comm_, &nccl_async_err);
+  if (nccl_err != ncclSuccess) {
+    throw std::logic_error(std::string("ncclGetAsyncError failed: ") + ncclGetErrorString(nccl_err));
+  }
+
+  if (nccl_async_err != ncclSuccess) {
+    ncclCommAbort(*nccl_comm_);
+    throw std::logic_error(std::string("NCCL async error: ") + ncclGetErrorString(nccl_async_err));
+  }
+
+
+}
+
 void NCCLOpContext::PopulateNCCLCommStrategy(int& nccl_rank, int& nccl_size,
                                              Communicator& nccl_id_bcast_comm) {
   if (communicator_type_ == Communicator::GLOBAL) {
@@ -136,16 +153,27 @@ Status NCCLAllreduce::Execute(std::vector<TensorTableEntry>& entries,
     num_elements += e.tensor->shape().num_elements();
   }
 
+  if (response.prescale_factor() != 1.0) {
+    // Execute prescaling op
+    ScaleBuffer(response.prescale_factor(), entries, fused_input_data, buffer_data, num_elements);
+    fused_input_data = buffer_data; // for unfused, scale is done out of place
+  }
+
   // Do allreduce.
   auto nccl_result = ncclAllReduce(fused_input_data, buffer_data,
                                    (size_t) num_elements,
                                    GetNCCLDataType(first_entry.tensor), ncclSum,
                                    *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
-  nccl_context_->ErrorCheck("ncclAllReduce", nccl_result);
+  nccl_context_->ErrorCheck("ncclAllReduce", nccl_result, *nccl_op_context_.nccl_comm_);
   if (global_state_->timeline.Initialized()) {
     gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_ALLREDUCE, *gpu_op_context_.stream);
   }
 
+  if (response.postscale_factor() != 1.0) {
+    // Execute postscaling op
+    ScaleBuffer(response.postscale_factor(), entries, buffer_data, buffer_data, num_elements);
+  }
+
   // Copy memory out of the fusion buffer.
   if (entries.size() > 1) {
     MemcpyOutFusionBuffer(buffer_data, entries);
@@ -155,7 +183,7 @@ Status NCCLAllreduce::Execute(std::vector<TensorTableEntry>& entries,
     }
   }
 
-  return gpu_op_context_.FinalizeGPUQueue(entries);
+  return gpu_op_context_.FinalizeGPUQueue(entries, true, nccl_op_context_.error_check_callback_);
 }
 
 #if HAVE_MPI
@@ -198,6 +226,12 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
     num_elements += e.tensor->shape().num_elements();
   }
 
+  if (response.prescale_factor() != 1.0) {
+    // Execute prescaling op
+    ScaleBuffer(response.prescale_factor(), entries, fused_input_data, buffer_data, num_elements);
+    fused_input_data = buffer_data; // for unfused, scale is done out of place
+  }
+
   // Do allreduce.
   int element_size = mpi_context_->GetMPITypeSize(first_entry.tensor->dtype());
   int local_size = global_state_->controller->GetLocalSize();
@@ -266,7 +300,7 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
                                          (size_t) num_elements_per_rank,
                                          GetNCCLDataType(first_entry.tensor),
                                          ncclSum, *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
-    nccl_context_->ErrorCheck("ncclReduceScatter", nccl_result);
+    nccl_context_->ErrorCheck("ncclReduceScatter", nccl_result, *nccl_op_context_.nccl_comm_);
     if (global_state_->timeline.Initialized()) {
       gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_REDUCESCATTER, *gpu_op_context_.stream);
     }
@@ -280,7 +314,7 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
                                   (size_t) num_elements_remaining,
                                   GetNCCLDataType(first_entry.tensor), ncclSum,
                                   root_rank, *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
-    nccl_context_->ErrorCheck("ncclReduce", nccl_result);
+    nccl_context_->ErrorCheck("ncclReduce", nccl_result, *nccl_op_context_.nccl_comm_);
     if (global_state_->timeline.Initialized()) {
       gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_REDUCE, *gpu_op_context_.stream);
     }
@@ -292,7 +326,7 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
     gpu_op_context_.host_buffer = malloc(total_buffer_len);
 
     // Synchronize.
-    gpu_context_->WaitForEvents(gpu_op_context_.event_queue, entries, timeline);
+    gpu_context_->WaitForEvents(gpu_op_context_.event_queue, entries, timeline, nccl_op_context_.error_check_callback_);
 
     // According to https://docs.nvidia.com/cuda/cuda-runtime-api/
     // api-sync-behavior.html#api-sync-behavior__memcpy-async,
@@ -325,7 +359,8 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
                               ncclAllGather(buffer_data_at_rank_offset, buffer_data,
                                             (size_t) num_elements_per_rank,
                                             GetNCCLDataType(first_entry.tensor),
-                                            *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream));
+                                            *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream),
+                              *nccl_op_context_.nccl_comm_);
     if (global_state_->timeline.Initialized()) {
       gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_ALLGATHER, *gpu_op_context_.stream);
     }
@@ -335,12 +370,18 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
                               ncclBcast(buffer_data_remainder,
                                         (size_t) num_elements_remaining,
                                         GetNCCLDataType(first_entry.tensor), root_rank,
-                                        *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream));
+                                        *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream),
+                              *nccl_op_context_.nccl_comm_);
     if (global_state_->timeline.Initialized()) {
       gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_BCAST, *gpu_op_context_.stream);
     }
   }
 
+  if (response.postscale_factor() != 1.0) {
+    // Execute postscaling op
+    ScaleBuffer(response.postscale_factor(), entries, buffer_data, buffer_data, num_elements);
+  }
+
   // Copy memory out of the fusion buffer.
   if (entries.size() > 1) {
     MemcpyOutFusionBuffer(buffer_data, entries);
@@ -350,7 +391,7 @@ NCCLHierarchicalAllreduce::Execute(std::vector<TensorTableEntry>& entries,
     }
   }
 
-  return gpu_op_context_.FinalizeGPUQueue(entries);
+  return gpu_op_context_.FinalizeGPUQueue(entries, true, nccl_op_context_.error_check_callback_);
 }
 
 bool NCCLHierarchicalAllreduce::Enabled(const ParameterManager& param_manager,
@@ -387,12 +428,192 @@ Status NCCLBroadcast::Execute(std::vector<TensorTableEntry>& entries,
                                       e.tensor->shape().num_elements() *
                                       DataType_Size(e.tensor->dtype()),
                                       ncclChar, e.root_rank,
-                                      *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream));
+                                      *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream),
+                            *nccl_op_context_.nccl_comm_);
   if (global_state_->timeline.Initialized()) {
     gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_BCAST, *gpu_op_context_.stream);
   }
 
+  return gpu_op_context_.FinalizeGPUQueue(entries, true, nccl_op_context_.error_check_callback_);
+}
+
+Status NCCLAllgather::Execute(std::vector<TensorTableEntry>& entries,
+                                const Response& response) {
+  auto& first_entry = entries[0];
+
+  gpu_op_context_.InitGPU(entries);
+  nccl_op_context_.InitNCCLComm(entries, response.devices());
+  gpu_op_context_.InitGPUQueue(entries, response);
+
+  // Sizes of subcomponents of each entry from all ranks
+  auto** entry_component_sizes = new int64_t* [entries.size()];
+
+  // Offset of each subcomponent of every entry in the final buffer after
+  // allgatherv
+  auto** entry_component_offsets = new int64_t* [entries.size()];
+
+  int global_size = global_state_->controller->GetSize();
+  int global_rank = global_state_->controller->GetRank();
+  auto* recvcounts = new int[global_size]();
+  auto* displcmnts = new int[global_size]();
+
+  for (size_t ec = 0; ec < entries.size(); ++ec) {
+    entry_component_sizes[ec] = new int64_t[global_size]();
+    entry_component_offsets[ec] = new int64_t[global_size]();
+  }
+
+  global_state_->timeline.ActivityStartAll(entries, ALLOCATE_OUTPUT);
+  Status status = AllocateOutput(entries, response, entry_component_sizes, recvcounts);
+  if (!status.ok()) {
+    for (size_t ec = 0; ec < entries.size(); ++ec) {
+      delete[] entry_component_sizes[ec];
+      delete[] entry_component_offsets[ec];
+    }   
+    delete[] entry_component_sizes;
+    delete[] entry_component_offsets;
+    delete[] recvcounts;
+    delete[] displcmnts;
+    return status;
+  }
+  global_state_->timeline.ActivityEndAll(entries);
+
+  SetDisplacements(recvcounts, displcmnts);
+  SetEntryComponentOffsets(entries, entry_component_sizes, recvcounts, entry_component_offsets);
+
+  size_t element_size = DataType_Size(first_entry.tensor->dtype());
+
+  const void* fused_input_data;
+  void* buffer_data;
+
+  // Copy memory into the fusion buffer.
+  if (entries.size() > 1) {
+    MemcpyInFusionBuffer(entries, displcmnts, element_size, buffer_data);
+    fused_input_data = (uint8_t*)buffer_data + displcmnts[global_rank] * element_size;
+
+    if (global_state_->timeline.Initialized()) {
+      gpu_context_->RecordEvent(gpu_op_context_.event_queue, MEMCPY_IN_FUSION_BUFFER, *gpu_op_context_.stream);
+    }
+  } else {
+    fused_input_data = first_entry.tensor->data();
+    buffer_data = (void*) first_entry.output->data();
+  }
+
+  bool same_shape = true;
+  const auto& tensor_sizes = response.tensor_sizes();
+  for (size_t ec = 0; ec < entries.size(); ++ec) {
+    for (int rc = 1; rc < global_size; ++rc) {
+      if (tensor_sizes[ec * global_size + rc] != tensor_sizes[ec * global_size]) {
+        same_shape = false;
+      }
+    }
+  }
+
+  // Do allgather.
+  if (same_shape) {
+    auto nccl_result = ncclAllGather(fused_input_data, buffer_data,
+                                     recvcounts[0] * element_size,
+                                     ncclChar,
+                                     *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
+
+    nccl_context_->ErrorCheck("ncclAllGather", nccl_result, *nccl_op_context_.nccl_comm_);
+
+    if (global_state_->timeline.Initialized()) {
+      gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_ALLGATHER, *gpu_op_context_.stream);
+    }
+  } else {
+    nccl_context_->ErrorCheck("ncclGroupStart", ncclGroupStart(), *nccl_op_context_.nccl_comm_);
+    for (int rc = 0; rc < global_size; ++rc) {
+      void* new_buffer_data = (uint8_t*)buffer_data + displcmnts[rc] * element_size;
+      auto nccl_result = ncclBroadcast(fused_input_data, new_buffer_data,
+                                       recvcounts[rc] * element_size,
+                                       ncclChar, rc,
+                                       *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
+      nccl_context_->ErrorCheck("ncclBroadcast", nccl_result, *nccl_op_context_.nccl_comm_);
+    }
+    nccl_context_->ErrorCheck("ncclGroupEnd", ncclGroupEnd(), *nccl_op_context_.nccl_comm_);
+
+    if (global_state_->timeline.Initialized()) {
+      gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_BCAST, *gpu_op_context_.stream);
+    }
+  }
+
+  // Copy memory out of the fusion buffer.
+  if (entries.size() > 1) {
+    MemcpyOutFusionBuffer(entry_component_offsets, entry_component_sizes,
+                          buffer_data, element_size, entries);
+
+    if (global_state_->timeline.Initialized()) {
+      gpu_context_->RecordEvent(gpu_op_context_.event_queue, MEMCPY_OUT_FUSION_BUFFER, *gpu_op_context_.stream);
+    }
+  }
+
+  delete[] recvcounts;
+  delete[] displcmnts;
+
+  for (size_t ec = 0; ec < entries.size(); ++ec) {
+    delete[] entry_component_sizes[ec];
+    delete[] entry_component_offsets[ec];
+  }
+  delete[] entry_component_sizes;
+  delete[] entry_component_offsets;
+
+  return gpu_op_context_.FinalizeGPUQueue(entries, true, nccl_op_context_.error_check_callback_);
+}
+
+bool NCCLAllgather::Enabled(const ParameterManager& param_manager,
+                              const std::vector<TensorTableEntry>& entries,
+                              const Response& response) const {
+  return entries[0].device != CPU_DEVICE_ID;
+}
+
+Status NCCLAlltoall::Execute(std::vector<TensorTableEntry>& entries,
+                             const Response& response) {
+#ifdef NCCL_P2P_SUPPORTED
+  assert(entries.size() == 1);
+  auto e = entries[0];
+
+  gpu_op_context_.InitGPU(entries);
+  nccl_op_context_.InitNCCLComm(entries, response.devices());
+  gpu_op_context_.InitGPUQueue(entries, response);
+
+  std::vector<int32_t> sdispls, rdispls;
+  std::vector<int32_t> sendcounts, recvcounts;
+  Status status = PrepareOutputAndParams(e, sdispls, rdispls, sendcounts, recvcounts);
+  if (!status.ok()) {
+    return status;
+  }
+
+  auto world_size = global_state_->controller->GetSize();
+  const void* sendbuf = e.tensor->data();
+  void* buffer_data = (void*) e.output->data();
+
+  nccl_context_->ErrorCheck("ncclGroupStart", ncclGroupStart(), *nccl_op_context_.nccl_comm_);
+
+  for (int i = 0; i < world_size; ++i) {
+    auto nccl_result = ncclRecv((uint8_t*) e.output->data() + rdispls[i] * DataType_Size(e.tensor->dtype()),
+                                recvcounts[i] * DataType_Size(e.tensor->dtype()), ncclChar, i,
+                                *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
+    nccl_context_->ErrorCheck("ncclRecv", nccl_result, *nccl_op_context_.nccl_comm_);
+
+    nccl_result = ncclSend((uint8_t*) e.tensor->data() + sdispls[i] * DataType_Size(e.tensor->dtype()),
+                           sendcounts[i] * DataType_Size(e.tensor->dtype()), ncclChar, i,
+                           *nccl_op_context_.nccl_comm_, *gpu_op_context_.stream);
+    nccl_context_->ErrorCheck("ncclSend", nccl_result, *nccl_op_context_.nccl_comm_);
+  }
+  nccl_context_->ErrorCheck("ncclGroupEnd", ncclGroupEnd(), *nccl_op_context_.nccl_comm_);
+
+  if (global_state_->timeline.Initialized()) {
+    gpu_context_->RecordEvent(gpu_op_context_.event_queue, NCCL_ALLTOALL, *gpu_op_context_.stream);
+  }
+
   return gpu_op_context_.FinalizeGPUQueue(entries);
+#else
+  throw std::runtime_error("NCCLAlltoall requires NCCL version >= 2.7.0. If your NCCL installation cannot be updated "
+                           "and you installed with HOROVOD_GPU_OPERATIONS=NCCL, reinstall with only supported "
+                           "operations individually specified (i.e. HOROVOD_GPU_ALLREDUCE=NCCL HOROVOD_GPU_BROADCAST=NCCL "
+                           "HOROVOD_GPU_ALLGATHER=NCCL). Otherwise, exclude HOROVOD_GPU_ALLTOALL=NCCL from your "
+                           "installation command.");
+#endif
 }
 
 } // namespace common
