@@ -13,11 +13,10 @@
 # limitations under the License.
 # ==============================================================================
 
-from __future__ import absolute_import
-
 import horovod.spark.common._namedtuple_fix
 
 import contextlib
+import os
 
 import pyarrow as pa
 import numpy as np
@@ -25,14 +24,41 @@ import pyspark.sql.functions as f
 from pyspark.ml.linalg import DenseVector, SparseVector, Vector, VectorUDT
 from pyspark.sql.types import ArrayType, BinaryType, BooleanType, FloatType, DoubleType, \
     IntegerType, LongType, NullType, StringType
-from pyspark.sql.types import from_arrow_type
+try:
+    # Spark 3.0 moved to a pandas submodule
+    from pyspark.sql.pandas.types import from_arrow_type
+except ImportError:
+    from pyspark.sql.types import from_arrow_type
 
-from horovod.run.common.util import codec
+from horovod.runner.common.util import codec, host_hash as hh
 from horovod.spark.common import cache, constants
 
 _training_cache = cache.TrainingDataCache()
 
 
+def host_hash(salt=None):
+    """
+    Computes this host's host hash by invoking horovod.runner.common.util.host_hash.host_hash.
+
+    Consider environment variable CONTAINER_ID which is present when running Spark via YARN.
+    A YARN container does not share memory with other containers on the same host,
+    so it must be considered a `host` in the sense of the `host_hash`.
+
+    :param salt: extra information to include in the hash, ignores Falsy values
+    :return: host hash
+    """
+    # turn salt into an array of a single string if given
+    salt = [str(salt)] if salt else []
+
+    # We would violate resource allocation if we run all tasks of a host in one container.
+    # See [issues 1497](https://github.com/horovod/horovod/issues/1497) for details.
+    container = os.environ.get("CONTAINER_ID")
+    if container is not None:
+        salt.append(container)
+
+    return hh.host_hash(salt='-'.join(salt))
+
+
 def data_type_to_str(dtype):
     if dtype == VectorUDT or dtype == SparseVector or dtype == DenseVector:
         return 'Vector'
@@ -122,7 +148,7 @@ def data_type_to_numpy(dtype):
 
 
 def check_shape_compatibility(metadata, feature_columns, label_columns,
-                              input_shapes=None, output_shapes=None):
+                              input_shapes=None, output_shapes=None, label_shapes=None):
     # Check for model and input type incompatibility. Columns must have the same size
     # (total number of elements) of the corresponding inputs.
     feature_count = len(feature_columns)
@@ -145,24 +171,36 @@ def check_shape_compatibility(metadata, feature_columns, label_columns,
                     'model input at index {idx} with size {input}'
                     .format(col=col, feature=col_size, idx=idx, input=input_size))
 
-    if output_shapes is not None:
-        label_count = len(label_columns)
-        if label_count != len(output_shapes):
-            raise ValueError('Label column count {labels} must equal '
-                             'model outputs count {outputs}'
-                             .format(labels=label_count, outputs=len(output_shapes)))
+    label_count = len(label_columns)
+    if label_shapes is not None and label_count != len(label_shapes):
+        raise ValueError('Label column count {labels} must equal '
+                         'provided label shapes count {outputs}'
+                         .format(labels=label_count, outputs=len(label_shapes)))
 
-        for idx, col, output_shape in zip(range(label_count), label_columns, output_shapes):
+    if output_shapes is not None and label_count != len(output_shapes):
+        raise ValueError('Label column count {labels} must equal '
+                         'model outputs count {outputs}'
+                         .format(labels=label_count, outputs=len(output_shapes)))
+
+    def _check_label_cols_size(target_shapes, target_name):
+        for idx, col, target_shape in zip(range(label_count), label_columns, target_shapes):
             col_size = metadata[col]['shape']
             if col_size is None:
                 # When training directly on Parquet, we do not compute shape metadata
                 continue
 
-            output_size = abs(np.prod(output_shape))
-            if col_size != output_size:
+            target_size = abs(np.prod(target_shape))
+            if col_size != target_size:
                 raise ValueError('Label column \'{col}\' with size {label} must equal that of the '
-                                 'model output at index {idx} with size {output}'
-                                 .format(col=col, label=col_size, idx=idx, output=output_size))
+                                 '{target_name} shape at index {idx} with size {output}'
+                                 .format(col=col, label=col_size, idx=idx, output=target_size,
+                                         target_name=target_name))
+
+    if label_shapes is not None:
+        _check_label_cols_size(label_shapes, 'label')
+    elif output_shapes is not None:
+        # Check the label size against the model output shapes only if label_shapes is not provided.
+        _check_label_cols_size(output_shapes, 'model output')
 
 
 def _get_col_info(df):
@@ -189,11 +227,6 @@ def _get_col_info(df):
                 size = data_col.indices.shape[0]
             elif isinstance(data_col, list):
                 shape = size = len(data_col)
-            elif isinstance(data_col, type(None)):
-                # Python 2.7 compat: NoneType is not pickleable
-                # see: https://bugs.python.org/issue6477
-                dtype = NullType
-                shape = size = 1
             else:
                 shape = size = 1
             row_schema.append((col_name, ({dtype}, {shape}, {size})))
