@@ -16,30 +16,32 @@
 # ==============================================================================
 # pylint: disable=g-short-docstring-punctuation
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
+import os
+import warnings
 
 from horovod.common.util import check_extension, gpu_available
 
 check_extension('horovod.tensorflow', 'HOROVOD_WITH_TENSORFLOW', __file__, 'mpi_lib')
 
+from horovod.tensorflow import elastic
 from horovod.tensorflow.compression import Compression
-from horovod.tensorflow.mpi_ops import allgather, broadcast, _allreduce
+from horovod.tensorflow.functions import allgather_object, broadcast_object, broadcast_object_fn, broadcast_variables
+from horovod.tensorflow.mpi_ops import allgather, broadcast, _allreduce, alltoall
 from horovod.tensorflow.mpi_ops import init, shutdown
 from horovod.tensorflow.mpi_ops import size, local_size, rank, local_rank, is_homogeneous
+from horovod.tensorflow.mpi_ops import rank_op, local_rank_op, size_op, local_size_op
 from horovod.tensorflow.mpi_ops import mpi_threads_supported, mpi_enabled, mpi_built
 from horovod.tensorflow.mpi_ops import gloo_enabled, gloo_built
-from horovod.tensorflow.mpi_ops import nccl_built, ddl_built, ccl_built
+from horovod.tensorflow.mpi_ops import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
 from horovod.tensorflow.mpi_ops import Average, Sum, Adasum
 from horovod.tensorflow.mpi_ops import handle_average_backwards_compatibility, check_num_rank_power_of_2
-
 from horovod.tensorflow.util import _executing_eagerly, _make_subgraph, _cache
+from horovod.tensorflow.mpi_ops import join
+from horovod.tensorflow.sync_batch_norm import SyncBatchNormalization
 
 import tensorflow as tf
-import warnings
 
-# @DEKHTIARJonathan: Do not remove, this fixes issues: 
+# @DEKHTIARJonathan: Do not remove, this fixes issues:
 # - https://github.com/tensorflow/tensorflow/issues/38516
 # - https://github.com/tensorflow/tensorflow/issues/39894
 if tf.__version__.startswith('2.2.'):
@@ -48,7 +50,8 @@ if tf.__version__.startswith('2.2.'):
 
 
 def allreduce(tensor, average=None, device_dense='', device_sparse='',
-              compression=Compression.none, op=None):
+              compression=Compression.none, op=None,
+              prescale_factor=1.0, postscale_factor=1.0):
     """Perform an allreduce on a tf.Tensor or tf.IndexedSlices.
 
     This function performs a bandwidth-optimal ring allreduce on the input
@@ -65,22 +68,28 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
                 Use `op` instead. Will be removed in v0.21.0.
 
         device_dense: Device to be used for dense tensors. Uses GPU by default
-                      if Horovod was built with HOROVOD_GPU_ALLREDUCE.
+                      if Horovod was built with HOROVOD_GPU_OPERATIONS.
         device_sparse: Device to be used for sparse tensors. Uses GPU by default
-                       if Horovod was built with HOROVOD_GPU_ALLGATHER.
+                       if Horovod was built with HOROVOD_GPU_OPERATIONS.
         compression: Compression algorithm used to reduce the amount of data
                      sent and received by each worker node.  Defaults to not
                      using compression.
         op: The reduction operation to combine tensors across different ranks.
             Defaults to Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
 
     Returns:
         A tensor of the same shape and type as `tensor`, summed across all
         processes.
     """
     op = handle_average_backwards_compatibility(op, average)
-    # Averaging happens in framework code, so translate that to Sum for the actual call
-    true_op = Sum if op == Average else op
+
+    average_in_framework = False
+    if rocm_built():
+        # For ROCm, perform averaging at framework level
+        average_in_framework = op == Average or op == Adasum
+        op = Sum if op == Average else op
 
     if isinstance(tensor, tf.IndexedSlices):
         # TODO: Need to fix this to actuall call Adasum
@@ -89,7 +98,8 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
                                       'workaround please pass sparse_as_dense=True to DistributedOptimizer')
         with tf.device(device_sparse):
             # For IndexedSlices, do two allgathers instead of an allreduce.
-            horovod_size = tf.cast(size(), tensor.values.dtype)
+            horovod_size = tf.cast(size_op() if int(os.environ.get("HOROVOD_ELASTIC", 0)) else size(),
+                                   dtype=tensor.values.dtype)
             values = allgather(tensor.values)
             indices = allgather(tensor.indices)
 
@@ -100,9 +110,12 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
                                 dense_shape=tensor.dense_shape)
     else:
         with tf.device(device_dense):
-            horovod_size = tf.cast(size(), dtype=tensor.dtype)
+            horovod_size = tf.cast(size_op() if int(os.environ.get("HOROVOD_ELASTIC", 0)) else size(),
+                                   dtype=tensor.dtype)
             tensor_compressed, ctx = compression.compress(tensor)
-            summed_tensor_compressed = _allreduce(tensor_compressed, op=true_op)
+            summed_tensor_compressed = _allreduce(tensor_compressed, op=op,
+                                                  prescale_factor=prescale_factor,
+                                                  postscale_factor=postscale_factor)
             summed_tensor = compression.decompress(summed_tensor_compressed, ctx)
             if op == Adasum:
                 if 'CPU' not in tensor.device and gpu_available('tensorflow'):
@@ -113,57 +126,45 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
                         elif not check_num_rank_power_of_2(int(size() / local_size())):
                             raise NotImplementedError(
                                 'Running GPU Adasum with non-power of 2 nodes is not supported yet.')
-                        horovod_local_size = tf.cast(local_size(), dtype=tensor.dtype)
-                        new_tensor = summed_tensor / horovod_local_size
+                        if rocm_built():
+                            horovod_local_size = tf.cast(local_size_op() if int(os.environ.get("HOROVOD_ELASTIC", 0)) else local_size(),
+                                                         dtype=tensor.dtype)
+                            new_tensor = summed_tensor / horovod_local_size
+                        else:
+                            new_tensor = summed_tensor
                     else:
                         warnings.warn('Adasum reduction does not currently support GPU reduction using MPI. Tensors '
                                       'are copied to CPU memory instead. To use Adasum for GPU reduction, please '
-                                      'compile Horovod with HOROVOD_GPU_ALLREDUCE=NCCL.')
+                                      'compile Horovod with HOROVOD_GPU_OPERATIONS=NCCL.')
                         new_tensor = summed_tensor
                 else:
                     if not check_num_rank_power_of_2(size()):
                         raise NotImplementedError('Running Adasum with non-power of 2 ranks is not supported yet.')
                     new_tensor = summed_tensor
             else:
-                new_tensor = (summed_tensor / horovod_size) if op == Average else summed_tensor
+                if rocm_built():
+                    new_tensor = (summed_tensor / horovod_size) if average_in_framework else summed_tensor
+                else:
+                    new_tensor = summed_tensor
         return new_tensor
 
 
-@_cache
-def _make_broadcast_group_fn():
-    if _executing_eagerly():
-        # Eager mode will parallelize independent control flow
-        def broadcast_group(variables, root_rank):
-            for var in variables:
-                var.assign(broadcast(var, root_rank))
+def _allreduce_cond(tensor, *args, **kwargs):
+    def allreduce_fn():
+        return allreduce(tensor, *args, **kwargs)
 
-        return _make_subgraph(broadcast_group)
-    else:
-        # Graph mode requires an Op
-        def broadcast_group(variables, root_rank):
-            return tf.group(*[var.assign(broadcast(var, root_rank))
-                              for var in variables])
+    def id_fn():
+        return tensor
 
-        return broadcast_group
-
-
-def broadcast_variables(variables, root_rank):
-    """Broadcasts variables from root rank to all other processes.
-
-    Arguments:
-        variables: variables for broadcast
-        root_rank: rank of the process from which global variables will be broadcasted
-                   to all other processes.
-    """
-    broadcast_group = _make_broadcast_group_fn()
-    return broadcast_group(variables, root_rank)
+    return tf.cond((size_op() > 1) if int(os.environ.get("HOROVOD_ELASTIC", 0)) else tf.convert_to_tensor(size() > 1),
+                   allreduce_fn, id_fn)
 
 
 try:
-    _global_variables = tf.global_variables
+    _global_variables = tf.compat.v1.global_variables
 except AttributeError:
     try:
-        _global_variables = tf.compat.v1.global_variables
+        _global_variables = tf.global_variables
     except AttributeError:
         _global_variables = None
 
@@ -186,10 +187,10 @@ if _global_variables is not None:
         return broadcast_variables(_global_variables(), root_rank)
 
 try:
-    _get_default_graph = tf.get_default_graph
+    _get_default_graph = tf.compat.v1.get_default_graph
 except AttributeError:
     try:
-        _get_default_graph = tf.compat.v1.get_default_graph
+        _get_default_graph = tf.get_default_graph
     except AttributeError:
         _get_default_graph = None
 
@@ -222,7 +223,7 @@ if _SessionRunHook is not None and _get_default_graph is not None:
                 Rank that will send data, other ranks will receive data.
               device:
                 Device to be used for broadcasting. Uses GPU by default
-                if Horovod was built with HOROVOD_GPU_BROADCAST.
+                if Horovod was built with HOROVOD_GPU_OPERATIONS.
             """
             super(BroadcastGlobalVariablesHook, self).__init__()
             self.root_rank = root_rank
@@ -240,7 +241,16 @@ if _SessionRunHook is not None and _get_default_graph is not None:
 
 @_cache
 def _make_allreduce_grads_fn(name, device_dense, device_sparse,
-                             compression, sparse_as_dense, op):
+                             compression, sparse_as_dense, op, gradient_predivide_factor):
+    if op == Average:
+        # Split average operation across pre/postscale factors
+        # C++ backend will apply additional 1 / size() factor to postscale_factor for op == Average.
+        prescale_factor = 1.0 / gradient_predivide_factor
+        postscale_factor = gradient_predivide_factor
+    else:
+        prescale_factor = 1.0
+        postscale_factor = 1.0
+
     def allreduce_grads(grads):
         with tf.name_scope(name + "_Allreduce"):
             if sparse_as_dense:
@@ -248,11 +258,13 @@ def _make_allreduce_grads_fn(name, device_dense, device_sparse,
                          if grad is not None and isinstance(grad, tf.IndexedSlices)
                          else grad for grad in grads]
 
-            return [allreduce(grad,
+            return [_allreduce_cond(grad,
                               device_dense=device_dense,
                               device_sparse=device_sparse,
                               compression=compression,
-                              op=op)
+                              op=op,
+                              prescale_factor=prescale_factor,
+                              postscale_factor=postscale_factor)
                     if grad is not None else grad
                     for grad in grads]
 
@@ -280,14 +292,15 @@ if _LegacyOptimizer is not None:
 
         def __init__(self, optimizer, name=None, use_locking=False, device_dense='',
                     device_sparse='', compression=Compression.none,
-                    sparse_as_dense=False, op=Average):
+                    sparse_as_dense=False, op=Average, gradient_predivide_factor=1.0):
             if name is None:
                 name = "Distributed{}".format(type(optimizer).__name__)
             super(_DistributedOptimizer, self).__init__(name=name, use_locking=use_locking)
 
             self._optimizer = optimizer
             self._allreduce_grads = _make_allreduce_grads_fn(
-                name, device_dense, device_sparse, compression, sparse_as_dense, op)
+                name, device_dense, device_sparse, compression, sparse_as_dense, op,
+                gradient_predivide_factor)
 
         def compute_gradients(self, *args, **kwargs):
             """Compute gradients of all trainable variables.
@@ -298,12 +311,9 @@ if _LegacyOptimizer is not None:
             allreduce the gradients before returning them.
             """
             gradients = self._optimizer.compute_gradients(*args, **kwargs)
-            if size() > 1:
-                grads, vars = zip(*gradients)
-                avg_grads = self._allreduce_grads(grads)
-                return list(zip(avg_grads, vars))
-            else:
-                return gradients
+            grads, vars = zip(*gradients)
+            avg_grads = self._allreduce_grads(grads)
+            return list(zip(avg_grads, vars))
 
         def apply_gradients(self, *args, **kwargs):
             """Calls this same method on the underlying optimizer."""
@@ -344,15 +354,15 @@ if _LegacyOptimizer is not None:
                 initializer=tf.zeros_initializer)
             self._is_first_step = tf.cast(tf.math.equal(self._step_count, 0), dtype=tf.bool)
             self._is_comm_step  = tf.cast(tf.math.equal(self._step_count % self._backward_passes_per_step, self._backward_passes_per_step - 1), dtype=tf.bool)
-        
+
         def _apply_shared(self, var, get_update_op):
             start_slot = self._get_or_make_slot(var, "delta_start")
 
             # initialize start on the first step
-            assign_op = tf.cond(self._is_first_step, 
-                lambda: start_slot.assign(var, use_locking=self.use_locking).op, 
+            assign_op = tf.cond(self._is_first_step,
+                lambda: start_slot.assign(var, use_locking=self.use_locking).op,
                 tf.no_op)
-            
+
             with tf.control_dependencies([assign_op]):
                 update_op = get_update_op()
                 with tf.control_dependencies([update_op]):
@@ -369,7 +379,7 @@ if _LegacyOptimizer is not None:
                         new_start = start_slot.assign_add(global_delta, use_locking=self.use_locking)
                         # var = start
                         return var.assign(new_start, use_locking=self.use_locking).op
-                    
+
                     # if its a communication step, then apply logic above
                     # if its not a communication step then just have the underlying
                     # optimizer update the model parameters according to its logic
@@ -421,7 +431,7 @@ if _LegacyOptimizer is not None:
 def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='',
                          device_sparse='', compression=Compression.none,
                          sparse_as_dense=False, backward_passes_per_step=1,
-                         op=Average):
+                         op=Average, gradient_predivide_factor=1.0):
     """Construct a new DistributedOptimizer, which uses another optimizer
     under the hood for computing single-process gradient values and
     applying gradient updates after the gradient values have been combined
@@ -439,10 +449,10 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
         See Optimizer.__init__ for more info.
       device_dense:
         Device to be used for dense tensors. Uses GPU by default
-        if Horovod was built with HOROVOD_GPU_ALLREDUCE.
+        if Horovod was built with HOROVOD_GPU_OPERATIONS.
       device_sparse:
         Device to be used for sparse tensors. Uses GPU by default
-        if Horovod was built with HOROVOD_GPU_ALLGATHER.
+        if Horovod was built with HOROVOD_GPU_OPERATIONS.
       compression:
         Compression algorithm used during allreduce to reduce the amount
         of data sent during each parameter update step.  Defaults to
@@ -458,7 +468,18 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
       op:
         The reduction operation to use when combining gradients across
         different ranks.
+      gradient_predivide_factor:
+        If op == Average, gradient_predivide_factor splits the averaging
+        before and after the sum. Gradients are scaled by
+        1.0 / gradient_predivide_factor before the sum and
+        gradient_predivide_factor / size after the sum.
     """
+    if gradient_predivide_factor != 1.0:
+        if rocm_built():
+            raise ValueError('gradient_predivide_factor not supported yet with ROCm')
+        if op != Average:
+            raise ValueError('gradient_predivide_factor not supported with op != Average')
+
     if isinstance(optimizer, _LegacyOptimizer):
         if op == Adasum:
             return _DistributedAdasumOptimizer(optimizer, name, use_locking, device_dense,
@@ -468,7 +489,8 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
                 raise ValueError('backward_passes_per_step>1 is not supported yet with '
                                  'op != Adasum')
             return _DistributedOptimizer(optimizer, name, use_locking, device_dense,
-                                        device_sparse, compression, sparse_as_dense, op)
+                                        device_sparse, compression, sparse_as_dense, op,
+                                        gradient_predivide_factor)
     elif isinstance(optimizer, tf.keras.optimizers.Optimizer):
         if op == Adasum:
             raise ValueError('op == Adasum is not supported yet with Keras')
@@ -476,7 +498,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
             raise ValueError('backward_passes_per_step > 1 is not supported yet with Keras')
         import horovod.tensorflow.keras as hvd_k
         return hvd_k.DistributedOptimizer(optimizer, name, device_dense, device_sparse,
-                                          compression, sparse_as_dense)
+                                          compression, sparse_as_dense, gradient_predivide_factor)
     else:
         raise ValueError('Provided optimizer doesn\'t inherit from either legacy '
                          'TensorFlow or Keras optimizer: %s' % optimizer)
@@ -485,7 +507,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
 if hasattr(tf, 'GradientTape'):
     class _DistributedGradientTape(tf.GradientTape):
         def __init__(self, tape, device_dense, device_sparse, compression, sparse_as_dense, op,
-                     persistent=False, watch_accessed_variables=True):
+                     gradient_predivide_factor, persistent=False, watch_accessed_variables=True):
             if hasattr(tape, '_watch_accessed_variables'):
                 super(self.__class__, self).__init__(persistent, watch_accessed_variables)
             else:
@@ -494,19 +516,16 @@ if hasattr(tf, 'GradientTape'):
             self._tape = tape
             self._allreduce_grads = _make_allreduce_grads_fn(
                 'DistributedGradientTape', device_dense, device_sparse, compression,
-                sparse_as_dense, op)
+                sparse_as_dense, op, gradient_predivide_factor)
 
         def gradient(self, target, sources, output_gradients=None):
             gradients = super(self.__class__, self).gradient(target, sources, output_gradients)
-            if size() > 1:
-                return self._allreduce_grads(gradients)
-            else:
-                return gradients
+            return self._allreduce_grads(gradients)
 
 
     def DistributedGradientTape(gradtape, device_dense='', device_sparse='',
                                 compression=Compression.none, sparse_as_dense=False,
-                                op=Average):
+                                op=Average, gradient_predivide_factor=1.0):
         """A tape that wraps another tf.GradientTape, using an allreduce to
         combine gradient values before applying gradients to model weights.
 
@@ -515,10 +534,10 @@ if hasattr(tf, 'GradientTape'):
             GradientTape to use for computing gradients and applying updates.
           device_dense:
             Device to be used for dense tensors. Uses GPU by default
-            if Horovod was built with HOROVOD_GPU_ALLREDUCE.
+            if Horovod was built with HOROVOD_GPU_OPERATIONS.
           device_sparse:
             Device to be used for sparse tensors. Uses GPU by default
-            if Horovod was built with HOROVOD_GPU_ALLGATHER.
+            if Horovod was built with HOROVOD_GPU_OPERATIONS.
           compression:
             Compression algorithm used during allreduce to reduce the amount
             of data sent during each parameter update step.  Defaults to
@@ -530,13 +549,24 @@ if hasattr(tf, 'GradientTape'):
           op:
             The reduction operation to use when combining gradients across
             different ranks.
+          gradient_predivide_factor:
+            If op == Average, gradient_predivide_factor splits the averaging
+            before and after the sum. Gradients are scaled by
+            1.0 / gradient_predivide_factor before the sum and
+            gradient_predivide_factor / size after the sum.
         """
+        if gradient_predivide_factor != 1.0:
+            if rocm_built():
+                raise ValueError('gradient_predivide_factor not supported yet with ROCm')
+            if op != Average:
+                raise ValueError('gradient_predivide_factor not supported with op != Average')
+
         cls = type(gradtape.__class__.__name__, (gradtape.__class__,),
                    dict(_DistributedGradientTape.__dict__))
         if hasattr(gradtape, '_watch_accessed_variables'):
             return cls(gradtape._tape, device_dense, device_sparse, compression,
-                       sparse_as_dense, op, gradtape._persistent,
+                       sparse_as_dense, op, gradient_predivide_factor, gradtape._persistent,
                        gradtape._watch_accessed_variables)
         else:
             return cls(gradtape._tape, device_dense, device_sparse, compression,
-                       sparse_as_dense, op, gradtape._persistent)
+                       sparse_as_dense, op, gradient_predivide_factor, gradtape._persistent)
