@@ -17,13 +17,8 @@ import inspect
 
 import tensorflow as tf
 
-from distutils.version import LooseVersion
-if LooseVersion(tf.__version__) >= LooseVersion("1.4.0"):
-    from tensorflow import keras
-    from tensorflow.python.keras import backend as K
-else:
-    from tensorflow.contrib import keras
-    from tensorflow.contrib.keras import backend as K
+from tensorflow import keras
+from tensorflow.python.keras import backend as K
 
 from horovod.tensorflow import init
 from horovod.tensorflow import shutdown
@@ -33,11 +28,11 @@ from horovod.tensorflow import rank
 from horovod.tensorflow import local_rank
 from horovod.tensorflow import mpi_threads_supported, mpi_enabled, mpi_built
 from horovod.tensorflow import gloo_enabled, gloo_built
-from horovod.tensorflow import nccl_built, ddl_built, ccl_built
+from horovod.tensorflow import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
 from horovod.tensorflow import Compression
 
 import horovod._keras as _impl
-from horovod.tensorflow.keras import callbacks
+from horovod.tensorflow.keras import callbacks, elastic
 
 
 try:
@@ -52,7 +47,8 @@ except:
 def DistributedOptimizer(optimizer, name=None,
                          device_dense='', device_sparse='',
                          compression=Compression.none,
-                         sparse_as_dense=False):
+                         sparse_as_dense=False,
+                         gradient_predivide_factor=1.0):
     """
     An optimizer that wraps another keras.optimizers.Optimizer, using an allreduce to
     average gradient values before applying gradients to model weights.
@@ -63,19 +59,27 @@ def DistributedOptimizer(optimizer, name=None,
               gradients. Defaults to "Distributed" followed by the provided
               optimizer type.
         device_dense: Device to be used for dense tensors. Uses GPU by default
-                      if Horovod was build with HOROVOD_GPU_ALLREDUCE.
+                      if Horovod was build with HOROVOD_GPU_OPERATIONS.
         device_sparse: Device to be used for sparse tensors. Uses GPU by default
-                       if Horovod was build with HOROVOD_GPU_ALLGATHER.
+                       if Horovod was build with HOROVOD_GPU_OPERATIONS.
         compression: Compression algorithm used to reduce the amount of data
                      sent and received by each worker node.  Defaults to not
                      using compression.
         sparse_as_dense: Treat all sparse gradients as dense tensors.  This can
                          help improve performance and memory utilization if
                          the original sparse gradient has high density.
-                         Defaults to false.    """
+                         Defaults to false.
+        gradient_predivide_factor: gradient_predivide_factor splits the averaging
+                                   before and after the sum. Gradients are scaled by
+                                   1.0 / gradient_predivide_factor before the sum and
+                                   gradient_predivide_factor / size after the sum.
+    """
+    if gradient_predivide_factor != 1.0 and rocm_built():
+            raise ValueError('gradient_predivide_factor not supported yet with ROCm')
+
     return _impl.create_distributed_optimizer(keras, optimizer, name,
                                               device_dense, device_sparse, compression,
-                                              sparse_as_dense)
+                                              sparse_as_dense, gradient_predivide_factor)
 
 
 def broadcast_global_variables(root_rank):
@@ -88,7 +92,7 @@ def broadcast_global_variables(root_rank):
     return _impl.broadcast_global_variables(K, root_rank)
 
 
-def allreduce(value, name=None, average=True):
+def allreduce(value, name=None, average=True, prescale_factor=1.0, postscale_factor=1.0):
     """
     Perform an allreduce on a tensor-compatible value.
 
@@ -98,8 +102,10 @@ def allreduce(value, name=None, average=True):
         name: Optional name for the constants created by this operation.
         average: If True, computes the average over all ranks.
                  Otherwise, computes the sum over all ranks.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
     """
-    return _impl.allreduce(K, value, name, average)
+    return _impl.allreduce(K, value, name, average, prescale_factor, postscale_factor)
 
 
 def allgather(value, name=None):
