@@ -1,5 +1,6 @@
 # Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
 # Modifications copyright Microsoft
+# Modifications copyright (C) 2020, NVIDIA CORPORATION. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,38 +15,24 @@
 # limitations under the License.
 # ==============================================================================
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
-from distutils.version import LooseVersion
-
 # Load all the necessary PyTorch C types.
 import torch
 
 import warnings
 
-# PyTorch v2 API starts with 1.0.0 (including nightly builds)
-_v2_api = LooseVersion(torch.__version__) >= LooseVersion('1.0.0')
-if _v2_api:
-    from horovod.torch import mpi_lib_v2 as mpi_lib
-    from horovod.common.basics import HorovodBasics as _HorovodBasics
-    _NULL = ""
-    _basics = _HorovodBasics(__file__, 'mpi_lib_v2')
-else:
-    from horovod.torch import mpi_lib_impl
-    from horovod.torch import mpi_lib
-    from horovod.common.basics import HorovodBasics as _HorovodBasics
-    _NULL = mpi_lib._ffi.NULL
-    _basics = _HorovodBasics(__file__, 'mpi_lib_impl', '_mpi_lib_impl')
+from horovod.torch import mpi_lib_v2 as mpi_lib
+from horovod.common.basics import HorovodBasics as _HorovodBasics
+_NULL = ""
+_basics = _HorovodBasics(__file__, 'mpi_lib_v2')
 
+from horovod.common.exceptions import HorovodInternalError
 from horovod.common.util import get_average_backwards_compatibility_fun, gpu_available, num_rank_is_power_2
 
 from horovod.torch.compression import Compression
 
 # import basic methods
 init = _basics.init
-shutdown = _basics.shutdown
+is_initialized = _basics.is_initialized
 size = _basics.size
 local_size = _basics.local_size
 rank = _basics.rank
@@ -58,6 +45,11 @@ gloo_built = _basics.gloo_built
 nccl_built = _basics.nccl_built
 ddl_built = _basics.ddl_built
 ccl_built = _basics.ccl_built
+cuda_built = _basics.cuda_built
+rocm_built = _basics.rocm_built
+def shutdown(*args, **kwargs):
+    mpi_lib.horovod_torch_reset()
+    return _basics.shutdown(*args, **kwargs)
 
 # import reduction op values
 Average = _basics.Average
@@ -74,9 +66,6 @@ handle_average_backwards_compatibility = get_average_backwards_compatibility_fun
 # before the operation is finished.
 _handle_map = {}
 
-# Only support fp16 allreduce for PyTorch versions using v2 API.
-_fp16_supported = _v2_api
-
 
 def _check_function(function_factory, tensor):
     function = function_factory(tensor)
@@ -91,15 +80,16 @@ def _allreduce_function_factory(tensor):
     return 'horovod_torch_allreduce_async_' + tensor.type().replace('.', '_')
 
 
-def _allreduce_async(tensor, output, name, op):
-    if tensor.dtype == torch.float16 and not _fp16_supported:
-        raise NotImplementedError(
-            'float16 allreduce is not supported for PyTorch version {} < 1.0.0'
-            .format(torch.__version__))
-
+def _allreduce_async(tensor, output, name, op, prescale_factor, postscale_factor):
     # Set the divisor for reduced gradients to average when necessary
     if op == Average:
-        divisor = size()
+        if rocm_built():
+            # For ROCm, perform averaging at framework level
+            divisor = size()
+            op = Sum
+        else:
+            divisor = 1
+
     elif op == Adasum:
         if tensor.device.type != 'cpu' and gpu_available('torch'):
             if nccl_built():
@@ -107,11 +97,15 @@ def _allreduce_async(tensor, output, name, op):
                     raise NotImplementedError('Running GPU Adasum on heterogeneous cluster is not supported yet.')
                 elif not num_rank_is_power_2(int(size() / local_size())):
                     raise NotImplementedError('Running GPU Adasum with non-power of 2 nodes is not supported yet.')
-                divisor = local_size()
+                if rocm_built():
+                    # For ROCm, perform averaging at framework level
+                    divisor = local_size()
+                else:
+                    divisor = 1
             else:
                 warnings.warn('Adasum reduction does not currently support GPU reduction using MPI. Tensors are '
                               'copied to CPU memory instead. To use Adasum for GPU reduction, please compile Horovod '
-                              'with HOROVOD_GPU_ALLREDUCE=NCCL.')
+                              'with HOROVOD_GPU_OPERATIONS=NCCL.')
                 divisor = 1
         else:
             if not num_rank_is_power_2(size()):
@@ -119,17 +113,20 @@ def _allreduce_async(tensor, output, name, op):
             divisor = 1
     else:
         divisor = 1
-    # Averaging happens in framework code, so translate that to Sum for the actual call
-    true_op = Sum if op == Average else op
 
     function = _check_function(_allreduce_function_factory, tensor)
-    handle = getattr(mpi_lib, function)(tensor, output, divisor,
-                                        name.encode() if name is not None else _NULL, true_op)
+    try:
+        handle = getattr(mpi_lib, function)(tensor, output, divisor,
+                                            name.encode() if name is not None else _NULL, op,
+                                            prescale_factor, postscale_factor)
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
     _handle_map[handle] = (tensor, output)
     return handle
 
 
-def allreduce_async(tensor, average=None, name=None, op=None):
+def allreduce_async(tensor, average=None, name=None, op=None,
+                    prescale_factor=1.0, postscale_factor=1.0):
     """
     A function that performs asynchronous averaging or summation of the input tensor
     over all the Horovod processes. The input tensor is not modified.
@@ -149,6 +146,8 @@ def allreduce_async(tensor, average=None, name=None, op=None):
         name: A name of the reduction operation.
         op: The reduction operation to combine tensors across different 
                    ranks. Defaults to Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
 
     Returns:
         A handle to the allreduce operation that can be used with `poll()` or
@@ -156,25 +155,30 @@ def allreduce_async(tensor, average=None, name=None, op=None):
     """
     op = handle_average_backwards_compatibility(op, average)
     output = tensor.new(tensor.shape)
-    return _allreduce_async(tensor, output, name, op)
+    return _allreduce_async(tensor, output, name, op, prescale_factor, postscale_factor)
 
 
 class HorovodAllreduce(torch.autograd.Function):
     """An autograd function that performs allreduce on a tensor."""
 
     @staticmethod
-    def forward(ctx, tensor, average, name, op):
+    def forward(ctx, tensor, average, name, op, prescale_factor, postscale_factor):
         ctx.average = average
         ctx.op = op
-        handle = allreduce_async(tensor, average, name, op)
+        ctx.prescale_factor = prescale_factor
+        ctx.postscale_factor = postscale_factor
+        handle = allreduce_async(tensor, average, name, op, prescale_factor, postscale_factor)
         return synchronize(handle)
 
     @staticmethod
     def backward(ctx, grad_output):
-        return allreduce(grad_output, average=ctx.average, op=ctx.op), None, None, None
+        return allreduce(grad_output, average=ctx.average, op=ctx.op,
+                         prescale_factor=ctx.prescale_factor,
+                         postscale_factor=ctx.postscale_factor), None, None, None, None, None
 
 
-def allreduce(tensor, average=None, name=None, compression=Compression.none, op=None):
+def allreduce(tensor, average=None, name=None, compression=Compression.none, op=None,
+              prescale_factor=1.0, postscale_factor=1.0):
     """
     A function that performs averaging or summation of the input tensor over all the
     Horovod processes. The input tensor is not modified.
@@ -201,17 +205,21 @@ def allreduce(tensor, average=None, name=None, compression=Compression.none, op=
                      not using compression.
         op: The reduction operation to combine tensors across different ranks. Defaults
             to Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
 
     Returns:
         A tensor of the same shape and type as `tensor`, averaged or summed across all
         processes.
     """
     tensor_compressed, ctx = compression.compress(tensor)
-    summed_tensor_compressed = HorovodAllreduce.apply(tensor_compressed, average, name, op)
+    summed_tensor_compressed = HorovodAllreduce.apply(tensor_compressed, average, name, op,
+                                                      prescale_factor, postscale_factor)
     return compression.decompress(summed_tensor_compressed, ctx)
 
 
-def allreduce_async_(tensor, average=None, name=None, op=None):
+def allreduce_async_(tensor, average=None, name=None, op=None,
+                     prescale_factor=1.0, postscale_factor=1.0):
     """
     A function that performs asynchronous in-place averaging or summation of the input
     tensor over all the Horovod processes.
@@ -231,16 +239,19 @@ def allreduce_async_(tensor, average=None, name=None, op=None):
         name: A name of the reduction operation.
         op: The reduction operation to combine tensors across different ranks. Defaults to
             Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
 
     Returns:
         A handle to the allreduce operation that can be used with `poll()` or
         `synchronize()`.
     """
     op = handle_average_backwards_compatibility(op, average)
-    return _allreduce_async(tensor, tensor, name, op)
+    return _allreduce_async(tensor, tensor, name, op, prescale_factor, postscale_factor)
 
 
-def allreduce_(tensor, average=None, name=None, op=None):
+def allreduce_(tensor, average=None, name=None, op=None,
+               prescale_factor=1.0, postscale_factor=1.0):
     """
     A function that performs in-place averaging or summation of the input tensor over
     all the Horovod processes.
@@ -260,12 +271,14 @@ def allreduce_(tensor, average=None, name=None, op=None):
         name: A name of the reduction operation.
         op: The reduction operation to combine tensors across different ranks. Defaults to
             Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensor before allreduce.
+        postscale_factor: Multiplicative factor to scale tensor after allreduce.
 
     Returns:
         A tensor of the same shape and type as `tensor`, averaged or summed across all
         processes.
     """
-    handle = allreduce_async_(tensor, average, name, op)
+    handle = allreduce_async_(tensor, average, name, op, prescale_factor, postscale_factor)
     return synchronize(handle)
 
 
@@ -275,8 +288,11 @@ def _allgather_function_factory(tensor):
 
 def _allgather_async(tensor, output, name):
     function = _check_function(_allgather_function_factory, tensor)
-    handle = getattr(mpi_lib, function)(
-        tensor, output, name.encode() if name is not None else _NULL)
+    try:
+        handle = getattr(mpi_lib, function)(
+            tensor, output, name.encode() if name is not None else _NULL)
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
     _handle_map[handle] = (tensor, output)
     return handle
 
@@ -355,8 +371,11 @@ def _broadcast_function_factory(tensor):
 
 def _broadcast_async(tensor, output, root_rank, name):
     function = _check_function(_broadcast_function_factory, tensor)
-    handle = getattr(mpi_lib, function)(
-        tensor, output, root_rank, name.encode() if name is not None else _NULL)
+    try:
+        handle = getattr(mpi_lib, function)(
+            tensor, output, root_rank, name.encode() if name is not None else _NULL)
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
     _handle_map[handle] = (tensor, output)
     return handle
 
@@ -471,6 +490,102 @@ def broadcast_(tensor, root_rank, name=None):
     handle = broadcast_async_(tensor, root_rank, name)
     return synchronize(handle)
 
+def _alltoall_function_factory(tensor):
+    return 'horovod_torch_alltoall_async_' + tensor.type().replace('.', '_')
+
+def _alltoall_async(tensor, splits, output, name):
+    if splits is None:
+        # If splits not provided, create empty tensor as placeholder
+        splits = torch.tensor([], dtype=torch.int32, device='cpu')
+    elif not isinstance(splits, torch.Tensor):
+        splits = torch.tensor(splits, dtype=torch.int32, device='cpu')
+
+    function = _check_function(_alltoall_function_factory, tensor)
+    try:
+        handle = getattr(mpi_lib, function)(
+            tensor, splits, output, name.encode() if name is not None else _NULL)
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
+    _handle_map[handle] = (tensor, splits, output)
+    return handle
+
+
+def alltoall_async(tensor, splits=None, name=None):
+    """
+    A function that scatters slices of the input tensor to all other Horovod processes
+    and returns a tensor of gathered slices from all other Horovod processes. The input
+    tensor is not modified.
+
+    The slicing is done on the first dimension, so the input tensors on
+    the different processes must have the same rank and shape, except for the
+    first dimension, which is allowed to be different.
+
+    Arguments:
+        tensor: A tensor to distribute with alltoall.
+        splits: A tensor of integers in rank order describing how many
+                elements in `tensor` to send to each worker.  Splitting is
+                applied along the first dimension of `tensor`. If `splits` is
+                not provided, the first dimension is split equally by the
+                number of Horovod processes.
+        name: A name of the alltoall operation.
+
+    Returns:
+        A handle to the alltoall operation that can be used with `poll()` or
+        `synchronize()`.
+    """
+    output = tensor.new()
+    return _alltoall_async(tensor, splits, output, name)
+
+
+class HorovodAlltoall(torch.autograd.Function):
+    """An autograd function that performs alltoall on a tensor."""
+
+    @staticmethod
+    def forward(ctx, tensor, splits, name):
+        ctx.tensor = tensor
+        ctx.splits = splits
+        handle = alltoall_async(tensor, splits, name)
+        return synchronize(handle)
+
+    @staticmethod
+    def backward(ctx, grad_output):
+        recvsplits = None
+        if ctx.splits is not None:
+            recvsplits = alltoall(ctx.splits, splits=torch.ones(size(), dtype=torch.int32, device='cpu'))
+        else:
+            splits_equal = torch.ones(size(), dtype=torch.int32, device='cpu') * (ctx.tensor.size()[0] // size())
+            recvsplits = alltoall(splits_equal, splits=torch.ones(size(), dtype=torch.int32, device='cpu'))
+        return alltoall(grad_output, splits=recvsplits), None, None
+
+
+def alltoall(tensor, splits=None, name=None):
+    """
+    A function that scatters slices of the input tensor to all other Horovod processes
+    and returns a tensor of gathered slices from all other Horovod processes. The input
+    tensor is not modified.
+
+    The slicing is done on the first dimension, so the input tensors on
+    the different processes must have the same rank and shape, except for the
+    first dimension, which is allowed to be different.
+
+    This acts as a thin wrapper around an autograd function.  If your input
+    tensor requires gradients, then callings this function will allow gradients
+    to be computed and backpropagated.
+
+    Arguments:
+        tensor: A tensor to distribute with alltoall.
+        splits: A tensor of integers in rank order describing how many
+                elements in `tensor` to send to each worker.  Splitting is
+                applied along the first dimension of `tensor`. If `splits` is
+                not provided, the first dimension is split equally by the
+                number of Horovod processes.
+        name: A name of the alltoall operation.
+
+    Returns:
+        A tensor containing the gathered tensor data from all workers.
+    """
+    return HorovodAlltoall.apply(tensor, splits, name)
+
 
 def poll(handle):
     """
@@ -502,9 +617,13 @@ def synchronize(handle):
     """
     if handle not in _handle_map:
         return
-    mpi_lib.horovod_torch_wait_and_clear(handle)
-    _, output = _handle_map.pop(handle)
-    return output
+
+    try:
+        mpi_lib.horovod_torch_wait_and_clear(handle)
+        output = _handle_map.pop(handle)[-1]
+        return output
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
 
 
 def join(device=-1):
@@ -519,6 +638,7 @@ def join(device=-1):
     Returns:
         Id of the rank that joined last.
     """
-    if not _v2_api:
-        raise NotImplementedError("Join Op is not supported for PyTorch < 1.0")
-    return mpi_lib.horovod_torch_join(device)
+    try:
+        return mpi_lib.horovod_torch_join(device)
+    except RuntimeError as e:
+        raise HorovodInternalError(e)
