@@ -13,21 +13,20 @@
 # limitations under the License.
 # ==============================================================================
 
-from __future__ import absolute_import
-
 import collections
+from distutils.version import LooseVersion
+import logging
 import warnings
 
 import mock
 import numpy as np
-import sys
+import pytest
 import tensorflow as tf
 
 import pyspark.sql.types as T
 from pyspark.ml.linalg import DenseVector, SparseVector
 from pyspark.sql.functions import udf
 
-from horovod.run.runner import is_gloo_used
 import horovod.spark.keras as hvd
 from horovod.spark.common import constants, util
 from horovod.spark.keras import remote
@@ -71,12 +70,10 @@ def get_mock_fit_fn():
 class SparkKerasTests(tf.test.TestCase):
     def __init__(self, *args, **kwargs):
         super(SparkKerasTests, self).__init__(*args, **kwargs)
+        logging.getLogger('py4j.java_gateway').setLevel(logging.INFO)
         warnings.simplefilter('module')
 
     def test_fit_model(self):
-        if sys.version_info < (3, 0, 0) and is_gloo_used():
-            self.skipTest('Horovod on Spark over Gloo only supported on Python3')
-
         model = create_xor_model()
         optimizer = tf.keras.optimizers.SGD(lr=0.1)
         loss = 'binary_crossentropy'
@@ -105,9 +102,6 @@ class SparkKerasTests(tf.test.TestCase):
                 assert pred.dtype == np.float32
 
     def test_fit_model_multiclass(self):
-        if sys.version_info < (3, 0, 0) and is_gloo_used():
-            self.skipTest('Horovod on Spark over Gloo only supported on Python3')
-
         model = create_mnist_model()
         optimizer = tf.keras.optimizers.Adadelta(1.0)
         loss = tf.keras.losses.categorical_crossentropy
@@ -229,6 +223,103 @@ class SparkKerasTests(tf.test.TestCase):
                     predictions = transformer.transform(df)
                 assert predictions.count() == df.count()
 
+    @mock.patch('horovod.spark.keras.remote._pin_gpu_fn')
+    @mock.patch('horovod.spark.keras.util.TFKerasUtil.fit_fn')
+    def test_keras_model_checkpoint_callback(self, mock_fit_fn, mock_pin_gpu_fn):
+        from horovod.tensorflow.keras.callbacks import BestModelCheckpoint
+
+        def _get_mock_fit_fn(checkpoint_callback_provided):
+            def fit(model, train_data, val_data, steps_per_epoch, validation_steps, callbacks,
+                    verbose):
+                returned_model_checkpoint_present = False
+                model_checkpoint_present = False
+                for callback in callbacks:
+                    callback.set_model(model)
+                    if checkpoint_callback_provided:
+                        callback.on_epoch_end(0, logs={'binary_crossentropy': 0.3})
+                    else:
+                        callback.on_epoch_end(0, logs={'binary_crossentropy': 0.3})
+
+                    if checkpoint_callback_provided and isinstance(callback, BestModelCheckpoint):
+                        self.assertIsNotNone(callback.filepath)
+                        self.assertTrue(callback.save_best_only)
+                        self.assertEqual(callback.monitor, 'binary_crossentropy')
+                        returned_model_checkpoint_present = True
+
+                    if not checkpoint_callback_provided and isinstance(callback, tf.keras.callbacks.ModelCheckpoint):
+                        self.assertFalse(callback.save_best_only)
+                        self.assertFalse(callback.save_best_only)
+                        self.assertEqual(callback.monitor, 'val_loss')
+                        model_checkpoint_present = True
+
+                if checkpoint_callback_provided:
+                    self.assertTrue(returned_model_checkpoint_present)
+                    self.assertFalse(model_checkpoint_present)
+                else:
+                    self.assertFalse(returned_model_checkpoint_present)
+                    self.assertTrue(model_checkpoint_present)
+
+                return mock.Mock()
+
+            return fit
+
+        mock_pin_gpu_fn.return_value = mock.Mock()
+
+        with spark_session('test_keras_model_chekcpoint_callbacks') as spark:
+            df = create_xor_data(spark)
+
+            backend = CallbackBackend()
+            with local_store() as store:
+                store.get_train_data_path = lambda v=None: store._train_path
+                store.get_val_data_path = lambda v=None: store._val_path
+
+                with util.prepare_data(backend.num_processes(),
+                                       store,
+                                       df,
+                                       feature_columns=['features'],
+                                       label_columns=['y']):
+                    model = create_xor_model()
+                    optimizer = tf.keras.optimizers.SGD(lr=0.1)
+                    loss = 'binary_crossentropy'
+
+                    # Test when the checkpoint callback is not set, the correct one is created
+                    mock_fit_fn.return_value = _get_mock_fit_fn(checkpoint_callback_provided=False)
+                    est = hvd.KerasEstimator(
+                        backend=backend,
+                        store=store,
+                        model=model,
+                        optimizer=optimizer,
+                        loss=loss,
+                        feature_cols=['features'],
+                        label_cols=['y'],
+                        batch_size=1,
+                        epochs=3,
+                        verbose=2)
+
+                    transformer = est.fit_on_parquet()
+                    predictions = transformer.transform(df)
+                    assert predictions.count() == df.count()
+
+                    # Test if checkpoint call back is correctly set to the model
+                    mock_fit_fn.return_value = _get_mock_fit_fn(checkpoint_callback_provided=True)
+                    checkpoint_callback = BestModelCheckpoint(monitor='binary_crossentropy')
+                    est = hvd.KerasEstimator(
+                        backend=backend,
+                        store=store,
+                        model=model,
+                        optimizer=optimizer,
+                        loss=loss,
+                        feature_cols=['features'],
+                        label_cols=['y'],
+                        batch_size=1,
+                        epochs=3,
+                        verbose=2,
+                        checkpoint_callback=checkpoint_callback)
+
+                    transformer = est.fit_on_parquet()
+                    predictions = transformer.transform(df)
+                    assert predictions.count() == df.count()
+
     @mock.patch('horovod.spark.keras.estimator.remote.RemoteTrainer')
     def test_model_serialization(self, mock_remote_trainer):
         model = create_xor_model()
