@@ -1,7 +1,3 @@
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import horovod.torch as hvd
 from horovod.common.util import env
 import torch
